// alloc 1024 memory blocks and test for memory bit errors
//   free good memory blocks (for other applications)
//   block bad memory blocks by sleeping forever after freeing good ones
//     == blacklist bad memory block at runtime
// if some bits are bad, 0.1% of memory will be occupied and rest can be used
// alternativ: linux badram/badmem kernel-option
//  EDAC = Error Detection And Correction
/*  grep+copy-friendly:
 gcc -Wall -O4 -fopenmp -DVSIZE=16 -o mem_badram mem_badram.c
 gcc -Wall -O2 -fopenmp -static    -o mem_badram mem_badram.c\
   -funroll-loops -ftree-vectorize -march=native # gcc -fopenmp gcc4.2++
 # intel-cc: icc -openmp
 GOMP_CPU_AFFINITY="0-3" OMP_NUM_THREADS=4 ./mem_badram $[7*1024] -l8 # 7GB
 GOMP_CPU_AFFINITY="0-3" OMP_NUM_THREADS=4 \
    srun -p short -t 50 -N 1-2 -l bash -c "env|grep OMP"
 # on linux check
 cat  /proc/sys/kernel/numa_balancing      # 0 = best_perf
 cat  /proc/sys/kernel/randomize_va_space  # 0 = best_perf
 cat  /proc/sys/vm/overcommit_memory       # 2 = no_overcommit (no OOM-panic)
 sync; echo 3 > /proc/sys/vm/drop_caches
 cat /sys/devices/system/edac/mc/poll_msec # CentOS5 k2.6.18 1000
 # verify: numastat # before + after malloc to see differences
 #   numactl --hardware # to see allocation per node
*/
// usage: mem_badram [<megabyte=1000>] [-l<nloops=100>] [-v<verbose=0>]
// HELP: use cpu affinity to check parts of numa memory on sockets/CPUs
// HELP: disable ECC at bios may be helpful identifying bad ram faster
// HELP: BIOS Data Scrambling
//   see https://en.m.wikipedia.org/wiki/Memory_controller + [6]=NXP+[7]
//  The memory controller incorporates a DDR3 Data Scrambling feature to
//  minimize the impact of excessive di/dt on the platform DDR3 VRs due to
//  successive 1s and 0s on the data bus.  Past experience has demonstrated
//  that traffic on the data bus is not random and can have energy
//  concentrated at specific spectral harmonics creating high di/dt that is
//  generally limited by data patterns that excite resonance between the
//  package inductance and on-die capacitances.  As a result, the memory
//  controller uses a data scrambling feature to create pseudo-random
//  patterns on the DDR3 data bus to reduce the impact of any excessive
//  di/dt.
// GOMP_CPU_AFFINITY="0-15"  # works as expected (gomp-1,comp2)
// OMP_PROC_BIND=TRUE # thread bounding, thread affinity (dont work!?)
// OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./mem_badram 7168 # 7GB test
// ./mem_badram 31000  # 31GB test, argument1 is physical memory in MB
//  OMP_NUM_THREADS=16 srun -p sw01_short -n 1 -t 00:10:00 ./mem_badram 31000 # /core
// for x in $(find /sys/devices/system/edac/ -name \*panic_on_ue);do
//   echo $x $(cat $x);
//   echo 0 > $x; done   # HELP: disable kernel panic on uncorrectable errors
//       # if you have UE-errors which panic your system, only some CE errors
//       #   may reported, so disabling panic_on_ue will detect more errors
// ToDo: a detect, but do not correct ECC mode would be a big advantage
//       Soft-ECC would be much more flexible than hardware ECC (or FPGA)
// ToDo: detect cpu-temp-sensor and output as power sensor (speed depends?)
//       /sys/devices/platform/coretemp.0/hwmon/hwmon5/temp1_input
//       t14 pk=76-80C idle=36-38C 25K/s..32K/s pwr=4.2W..23W(BAT=23-90C)
//       t14 +30K/1s or srate=1s! (sampling 0.1s shows only 1s-changes)
//       + maybe core voltage , sampling rate?
//  --task-mem=450 ** OpenMP fatal error: cannot allocate critical lock
//
// 2020-07-04 better performance for EpycCPUs 4*2mem_channels,
//   parallel for of blocks (old: within blocks) for more numa (epyc)
//   (slow) seq. initialization for speedup non-interleaved memory channels
//   2*(8ch-epyc) = 290GiB/s, 2*(4ch-haswell=44GiB/s), write/read=40-55%
//   performance matters if you have to search bad DIMMs by plugging
//   speedup 2..4 (now ca.80% of theoretical BW)
//   WebsearchMaxBW: K40c 70%(ECC) 64%(noECC),
//        2ch*DDR3-1600MHz=25.6GB/s write32bit=read8B(cacheline)+modify4B
//        non-temporal-write=__mm256_stream_si256(&a,const8B)=memset (not store)
//        + cld;rep stosq 20.6GB/s=80%(1t), 4*memset=22.1GB/s=86%
//   https://codearcana.com/posts/2013/05/18/achieving-maximum-memory-bandwidth.html
//       N455 2*memset=1*memset=rep_stosq=1.3GB/s no speedup lib-version? cpu?
//       hasvel rep_stosq ileave1:1 50GB/s (loop=37GB/s), not 100%r=88GB/s
//       ryzen rep_stosq 100%r
// 2020-07-09 trying asm rep stosq (near 100% on modern CPUs: PC16)
// 2020-06-18 precise output GiB/s (old=MB/s but was MiB/s)
//            fix bad factor 2 speed on ecc half-block-test
//            problem: multicore gives overlapping addresses for blocks
//            t100 non-ecc + non-interleave tests with bad DIMM n017.D1
// 2018-04-04 improvements for slurm-cluster +ECC-RAM health scan
//            add ompi-support (use slurm or ompi to start on many nodes)
// 2015-10-28 change OMP-loop-vars to signed (avoid pathcc.omp warn+crash)
// 2015-09-04 better avoid in cache-loops, correct speed if malloc-fail
// 2015-08-31 +openmp, fix use of lrnd, +output speed
// 2013-03-22 improved version
// 2011-07-06 cosmetic changes for output (more info)
// 2013 better use memspeed --test=8 or 16  ???
// 2022-06-21 fix multiple outputs for slurm multinode start
//
// ToDo20: noECC get physical mem-address and DIMM-No from virtual address
// ToDo18 w+r all, better partly wait longer (skip r/writes) log2-ranges
//   to have row-hammer like pattern .....xxxxxx...xxx...xxx..xx..xx..xx.x.x.x.x
// ToDo: disable ECC in BIOS for tests? Asus=yes(ibio,t100)
//       lsmod | grep edac # sb_edac edac_core
//       cat /sys/devices/system/edac/mc/mc[01]/[uc]e_*count|xargs
//       https://www.kernel.org/doc/Documentation/edac.txt
//   dmidecode 64bit = internal ECC only (if so)
//             72bit = external ECC on Board or CPU(+memctl) or FakeECC 
//    k2.6.16 /sys/devices/system/machinecheck/machinecheck0/bank* # mcelog?
// ToDo: spinpack: Immunity-aware programming, p.e. check sum matrix_elements
//     and compare every iteration (stored=memerror, recompute=compute_error?)
//
// before that do: echo 3 >/proc/sys/vm/drop_caches
//  cat /sys/devices/system/edac/mc/mc{0,1}/csrow{0,1}/{c,u}e_count|xargs
// for x in $(find /sys/* -name \*[cu]e_\*count);do echo $x $(cat $x);done
//  ## ch?_ce_count ce_noinfo_count ce_count ue_count
// use: dmidecode # to detect ECC-functionality
// GOMP_CPU_AFFINITY="0-15:1" 
// 2015-09x31_______2GB_rw1.5GB                 348MB/s 1t 2004 noECC PentiumM-1400MHz@1200MHz gcc451
// 2020-07loki05_2x2c8GB_3.6GB/c  3-6GiB/s 4t  2611MB/s 1t 2005 ECC 2x2cOpt275-2.2GHz dmi=6*2GB DualChannel-DDR-400=3.2GiB/s/ch CentOS5 page=0x441906+ofs358 = 2exe_mapped same maps! 2ch*4csrow
// 2018-06comp1_4x2c32GB_3.6GB/c   5.4GB/s 8t  1055MB/s 1t 2006 ECC-1defDIMM Opt885-2.6GHz+16*2GB-DDR-400MHz-gcc492vec read8t=15.1GB/s read1t=2.3GB/s
// 2020-07comp1_4x2c32GB_3.6GB/c  14.8GB/s 8t  2.26GB/s 1t 2006 ECC gcc4.9_meggie 4CPUs*(2ch*400MT/s*8B=6.4GB/s)=25.6GB/s 1t=70% w/r=81% 8t=58% w/r=55% c0+4m0=4.3GB/s c1+5m1 c2+6m2 k2.6.16
// 2015-09ibioN_2x2x1GB_rw900MB/c  2.8GB/s 4t  1030MB/s 1t 2006 ECC Opt2214-2.2GHz gcc412-O2+unroll+vect dmidecode=64bit-DIMMs board=SbECC (aber 9chips/DIMM) 2*2of4DIMMs gcc492+k8=2.5GB/s/4c bios.ecc=on memtest86+v420=2GB/s 4*singletask=4*720MB/s
// 2015-09ibioM_1x4x1GB_rw1.6GB/c  1.4GB/s 2t  1168MB/s 1t 2006 ECC Opt275-2.2GHz  gcc412-O2+unroll+vect dmidecode=72bit-DIMMs
// 2015-09mkautz12x2x2GB_rw500MB/c 876MB/s 6t   162MB/s 1t 2008 ECC m5k.gcc412 -O2 unroll no_openmp! 6t=6*single DDR2
// 2015-10mkautz12x2x2GB_rw500MB/c 6xxMB/s 6t   162MB/s 1t 2008 ECC m5k.gcc424 -O2 unroll +openmp t6schwankt=413-690MB/s! 2t=303MB/s 3t=420MB/s 4t=bad=290-305
// 2015-10mkautz12x2x2GB_rw500MB/c 780MB/s 6t   162MB/s 1t 2008 ECC m5k.pathcc -O2 unroll +openmp loop-var-ijk-must-be-signed!       3t=409MB/s 500MHz
// 2015-10asgard12x2x2GB_rw500MB/c 929MB/s 6t   203MB/s 1t 2008 ECC m5k.pathcc -O2 unroll +openmp                                    3t=505MB/s 680MHz
// 2015-10asgard12x2x2GB_rw500MB/c      -       202MB/s 1t 2008 ECC m5k.gcc412=gcc464 -O2 unroll 680MHz
// 2015-10asgard12x2x2GB_rw500MB/c 1.0GB/s 6t   198MB/s 1t 2008 ECC m5k.gcc464 -O2 unroll +openmp 680MHz mit/ohne bootstrap          3t=569MB/s 6t=6*1t-20%
// 2015-10asgard12x2x2GB_rw500MB/c 899GB/s 6t   170MB/s 1t 2008 ECC m5k.gcc492 -O2 unroll +openmp 680MHz --disable-bootstrap
// 2017-08kautz972x2x2GB_rw2000MB       -       140MB/s 1t 2009 ECC m5k.gcc412 -O2 500MHz DDR2-RAM
// 2017-08kautz972x2x2GB_rw400MB/c 701MB/s 6t   163MB/s 1t 2009 ECC m5k.pathcc -O2 500MHz +openmp
// 2016-09asgardsp_1x1GB_rw750MB/c      -  1t   874MB/s 1t 2008 Sempron3600+800MHz+DDR2-266MHz gcc412
// 2015-10mkautzsp_1x1GB_rw500MB/c      -       902MB/s 1t 2008 noECC?mc! Sempron3500+800MHz k8=sse2+3dnow
// 2015-09quantum1x4x2GB_rw1GB/c:  1.7GB/s 4t   1.2GB/s 1t 2008 ECC E5405-2GHz 228W-200W dmidecode=72bit-DIMMs DDR2 Multi-bit-ECC
// 2015-09quantum2x4x4GB_rw2GB/c:  2.3GB/s 8t   643MB/s 1t 2008 ECC E5450-3GHz 
// 2015-09jsPC_____1x2GB_rw1GB     1.2GB/s 2t   1.2GB/s 1t 2010 noECC E5400-2.7GHz@1.6GHz
// 2016-06jsPC_____2x2GB_rw3GB    1.33GB/s 2t  1.30GB/s 1t 2010 noECC E5400-2.7GHz@1.6GHz TODO memtest86+
// 2015-09einstein2x2GB_rw1.5GB    2.7GB/s 4t   2.4GB/s 1t 2010 ECC? 4c8t-i7-860-2.8GHz gcc44-32bit! no-sse4-used! dmidecode=72bit-DIMMs BoardECC=none
// 2015-09landau__6x4GB_rw12GB     5.3GB/s 4t   4.1GB/s 1t 2010 ECC? 4c*2HT-iCore_i7-950-3.06GHz-8MB SSE4.2 24GB dmidecode=2304/2244bits? board=noECC
// 2015-08comp2__4x4x4GB_rw3GB/c: 15.8GB/s 16t  2.0GB/s 1t 2011 ECC Opt6134-2.3GHz
// 2015-09comp2n_4x4x8GB_rw7GB/c: 16.5GB/s 16t  2.1GB/s 1t 2011 ECC Opt6134-2.3GHz.gcc492-O2+unroll+vect
// 2020-07comp2n_4x4x8GB_7.4GB/c  43.6GB/s 16t  4.1GB/s 1t 2011 ECC Opt6134-2.3GHz.gcc492 K10 mem-channels? 4/cpu? 2?*(DDR3-667=5.3GB/s)/CPU mc0-3 rank0-3 TODO
// 2015-09gibbs____4x4GB_rw14GB    5.9GB/s 4t   5.1GB/s 1t 2011 noECC? i5-2400-3.1GHz gcc412-O2+unroll+vect+omp hdc=30130h=3.4y dmidecode=64bit-DIMMs board=noECC
// 2015-09hpmini_2GB_1500MB:      1020MB/s 2t   990MB/s 1t 2012 noECC 6.5W-N455-1660MHz+gcc472-O2+unroll+vec BitError!
// 2020-07hpmini_2GB_900MB        4.05GB/s 2t  3.17GB/s 1t 2012 noECC 6.5W-N455-1660MHz+gcc49 1ch*SODIMM-2GB-DDR2-667MHz=5.3GB/s
// 2015-09meitner__4x8GB_rw16GB    5.6GB/s 4t   5.3GB/s 1t 2013 noECC? 4c8ht-i7-3770-Gen2-3410MHz-avx DDR3 dmidecode=64bit-DIMMs
// 2020-07x230t_1*4GB_1.2GB/c     10.9GB/s 2t  10.7GB/s 1t 2013? noECC 35W-2c-i5-3320M-2.6GHz-2ch*12.8GB/s 1*4GB-DDR3-1600MHz-SODIMM=12.8GB/s
// 2015-09dyson__1x4x8GB_rw7GB/c:  6.8GB/s 4t   6.0GB/s 1t 2014 noECC i5-4460-3.2GHz+gcc482-25.6GB/s 80W dmidecode=64bit-DIMMs
// 2015-09dyson__1x4x8GB_rw7GB/c:  5.9GB/s 4t   2.1GB/s 1t 2014 noECC i5-4460-800MHz+gcc482 49W
// 2015-09msiKS____2x4GB_rw4GB     5.6GB/s 4t   5.1GB/s 1t 2015 noECC i5-4210H-2.9GHz+gcc492-O3+unroll+vec
// 2015-12t100 2x(E5-2630v3-2.4GHz-8c) 2x(2DIMMs/ch 1866MHz*8B=14.9GB/s/ch 4ch=59GB/socket)
// 2015-12t100_2x8x16GB_14GB/c:   30.4GB/s 16t  5.1GB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc483 # O2=30.4GB/s O2+unroll+vect=30.5GB/s
// 2015-12t100_2x8x16GB_14GB/c:   30.3GB/s 16t  5.xGB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.3GB/s n151.2016-12-12
// 2015-12t100_2x8x16GB_11GB/c:   30.7GB/s 16t  5.2GB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.5GB/s n151.2016-12-12
// 2015-12t100_2x6x16GB_11GB/c:   22.8GB/s 16t  5.0GB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.5GB/s n152.2016-12-12
// 2015-12t100_2x8x16GB_11GB/c:   30.3GB/s 16t  5.xGB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.5GB/s n152.2016-12-15 nach replace bad DIMM
// 2018-03t100_32*8GB_7GiB/c:     29.7GB/s 32t  4.9GB/s 1t 2016 ECC gcc483.O2=20GB/s,r=63GB/s 16t=30.4GB/s    32*7=224GiB=229376MB
// 2018-03t100_32*8GB_5.5GiB/c:   29.7GB/s 32t  4.9GB/s 1t 2016 ECC gcc483.O2=20GB/s,r=63GB/s 16t=30.4GB/s  32*5.5=176GiB=180224MB
// 2018-03t100_16*16GB_11GiB/c:   30.2GB/s 16t  5.0GB/s 1t 2016 ECC gcc483.O2=20GB/s,r=63GB/s 16t=30.4GB/s  32*5.5=176GiB=180224MB
// 2020-06t100_16*16GB_8GB/c:     92.2GB/s 16t   # auto-interleaved 8t? 16t! omp
// 2020-06t100_16*16GB_8GB/c:     13.7GB/s 8t    # 1-way/ch 1-way/csrow  omp
// 2020-07t100_16*16GB_15.6GB/c   44.0GB/s 8t    # 1-way/ch 1-way/csrow c0-7 omp
// 2020-07t100_16*16GB_15.6GB/c   89.0GB/s 16t   # 1-way/ch 1-way/csrow c0-15 omp non-interleaved (+omp-threads+handloopAVX)
//   until 2020-05 here to slow measurement
// 2020-06t100_16*16GB_r15GB/c    48.2GB/s  8t 11.5GB/s 1t # cpu0 m0 only w=22.6GB/s ab30GB 1cpu8c=59GB/s*82%
// 2020-06t100_16*16GB_r15GB/c     7.8GB/s  8t  7.5GB/s 1t # cpu0 m1 only w=5.0GB/s 1cpu8c=59GB/s_max
// 2020-06t100_16*16GB_r15GB/c    95.3GB/s 16t 11.5GB/s 1t # 2cpus/node w=44.2GB/s ab30GB 1cpu8c=59GB/s*81% no-vec
// 2020-07t100_16*16GB_r15GB/c    80.4GB/s 16t 11.4GB/s 1t # i1:1=non-interleaved mem numa_balancing=0 n016+bad_blk=66GiB/s
// 2020-06jsPC16_2*8GB_r1GB/c:    26.4GB/s  8t 15.7GB/s 1t 2016 78%*34GB/s i7-6700-3.4GHz-4c8t 2DIMMs*8GB-DDR4-2133MT/s=17GB/s 2ch=34.1GB/s i7-6700-3.4GHz-4c8t w=50%r ab 1-4GB 2wmemset=13.4GB/s rep_stosq=25GiB/s(1st=28?)
// 2020-06jsPC16_2*8GB_6GB:       25.0GB/s  4t 16.5GB/s 1t 2016 78%*34GB/s i7-6700-3.4GHz-4c8t 2DIMMs*8GB-DDR4-2133MT/s=17GB/s 2ch=34.1GB/s w=50%r ab 1-4GB 2wmemset=13.4GB/s rep_stosq=25GiB/s(1st=28?) +_mm_prefetch
// 2020-07hpc18_16*16GB_230GB     283GiB/s 32t 20.1GB/s 1t 2018 2*(16c-epyc7351-2.4GHz-4*2ch) ecc-ddr4-2667MT/s*8=21.3GiB/s/ch 2*4*2ch=max340GiB/s w=121GB/s +prefetch256x8=283GiB/s prefetch128x8=290GiB/s no_prefetch=280GiB/s numa_balancing=0
// 2022-01hpc21 16*16GB_240GB     233GiB/s 32t 15.6GB/s 1t 2021 2*(16c-iX6326-2.9GHz-8ch) ecc-ddr4-3200MT/s*8=25.6GiB/s
// 2022-01hpc21 16*16GB_240GB     242GiB/s 32t 14.0GB/s 1t 2021 2*(16c-iX6326-2.9GHz-8ch) ecc-ddr4-3200MT/s*8=25.6GiB/s/ch
//   numademo 1GB memset # memset=5.8GB/s stream=7.3GB/s
// iLoop=   1.15 speed.r: 123.907 - 125.212 GiB/s (126000 MiB) # 3300MHz
// iLoop=   1.05 speed.r: 123.089 - 124.001 GiB/s (126000 MiB) # 3000MHz
// iLoop=   0.03 speed.r: 122.307 - 124.138 GiB/s (125000 MiB) # 2800MHz max
// iLoop=   0.08 speed.r: 121.204 - 122.242 GiB/s (126000 MiB) # 2400MHz
// iLoop=   0.06 speed.r: 117.764 - 119.245 GiB/s (126000 MiB) # 2000MHz ok
// iLoop=   0.11 speed.r: 114.275 - 115.812 GiB/s (126000 MiB) # 1600MHz ok
// iLoop=   0.06 speed.r: 107.571 - 109.493 GiB/s (126000 MiB) # 1200MHz
// iLoop=   0.06 speed.r:  87.326 -  88.108 GiB/s (126000 MiB) #  800MHz
// background-load nice -19 taskset -c 0 dd urandom 10GB sys=12% 38s -> 1465s+ 
//  for x in $(seq 32 63);do nice -19 taskset -c $x \...
//    dd if=/dev/urandom of=/dev/null bs=1M count=1000 & done;wait
//  ## 48s 22MB/s + c0-31_badram: 65-73s 14.4MB/s = 76% c32+load + 100% c0+mem
//  ##         280s c0+ badram_249GB/s+nice_urandom_3.7MB/s (share vcores)
// cpupower frequency-set --max  800000
// # write is slower because default cacheline-modify, read 64B +mod +write 64B
// t100: GOMP_CPU_AFFINITY="0-15" srun -p sw01_short -n 1 -t 00:10:00 bash -c
//        "OMP_NUM_THREADS=16 ./mem_badram 245760" # (16c*14GB/c=224GB)*1024MB/GB
//        # 2018-03 change to 0-31 !?
//  OMP_NUM_THREADS=16 GOMP_CPU_AFFINITY="0-15" numactl -l -C 0-15 ./mem_badram $[240*1024] # 30.5GB/s
//  OMP_NUM_THREADS=8 GOMP_CPU_AFFINITY="0-7"   numactl -l -C 0-7  ./mem_badram $[120*1024] # 15.3GB/s
//  OMP_NUM_THREADS=8 GOMP_CPU_AFFINITY="0-7" numactl -m 1 -C 0-7  ./mem_badram $[120*1024] #  3.0GB/s
//    check correct pinning on running HPC (high-mem) apps (health?)
//     + other 8 cores crosswise: 2*3064MB/s  check: ps -am -o spid,psr,cmd
//    test -N 3 OK, c129 = 9GB/s only (1st round, drop_caches?), c129.2nd = 22.5GB/s
//     3th = 24.6GB/s  ToDo: min/max nodes/cores
// cmp: https://www.techspot.com/news/62129-ddr3-vs-ddr4-raw-bandwidth-numbers.html
//   AIDA64 Read Bandwith? aus Bild (cpu/memctlr-abhaengig!)
//   read-DDR3-1866-4ch = 49GB/s = 59.7GB/s * 82%  write=33GB/s=67%r cp=43GB/s
//   read-DDR4-2133-4ch = 54GB/s = 68.3GB/s * 79%  write=47GB/s=87%r cp=52GB/s
//  DDR3 1600/1866/2133/2400/2666/2800MHz  t100=1866MHz(4*14.9GHz/cpu)
//  DDR4 2133/2400/2666/2800/3000/3200MHz  PC16=2133MHz(2*17GB/s*78%)
//    latencies 45(2800MHz)-80ns(1600MHz)
//
/* --- 2015-09hpmini_2GB_2012 bit-flip-error ---
 malloc 1024 MiB (Mi=2^20) long= 8 B = 64 bit (arg1)
 malloc 1024 blocks to 1024 kiB = 131072 * long (ki=2^10)
 starting, only every 128th block shown ...
 a=00007fb978db0010 = 130789 GB blk    0 (alloc+ini)
 a=00007fb9708a0010 = 130789 GB blk  128 (alloc+ini)
  ...
 a=00007fb948620010 = 130789 GB blk  768 (alloc+ini)
 a=00007fb9405a0010 = 130789 GB blk  896 (alloc+ini)
 a=00007fb978db0010 iLoop=00/16 ffffffffffffffff
 a=00007fb978db0010 iLoop=02/16 ffffffffffffffff
 loop=  0 speed(w+r):        887 MB/s (1500.000 MB)
 a=00007fb957361010 iLoop=00/16 9bc3679699ccb78c
 a=00007fb957361010 iLoop=02/16 9bc3679699ccb78c
 a=00007fb957361010 iLoop=04/16 9bc3679699ccb78c  # 1-to-0 bit 15
ERR bad data blk=734 lrnd= 9bc3679699ccb78c  sum=00008000 (j=4) on ?
ERR bad data   a[0001c603]=9bc3679699cc378c diff=00008000 (j=4)
ERR bad data blk=734 lrnd= d5176bb59daaa713  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=d5176bb59daa2713 diff=00008000 (j=2)
ERR bad data blk=734 lrnd= ffffffffffffffff  sum=00008000 (j=8) on ?
ERR bad data   a[0001c603]=ffffffffffff7fff diff=00008000 (j=8)
ERR bad data blk=734 lrnd= ee6f32189910bc72  sum=00008000 (j=10) on ?
ERR bad data   a[0001c603]=ee6f321899103c72 diff=00008000 (j=10)
ERR bad data blk=734 lrnd= ef6306d1f2128d9c  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=ef6306d1f2120d9c diff=00008000 (j=2)
ERR bad data blk=734 lrnd= 7fdcc2331befd79f  sum=00008000 (j=1) on ?
ERR bad data   a[0001c603]=7fdcc2331bef579f diff=00008000 (j=1)
ERR bad data blk=734 lrnd= be583b369497f865  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=be583b3694977865 diff=00008000 (j=2)
 haeufig, dann weniger, dann haeufiger bei parallelem 20MB-scan = 20kB blks
 heating-related? ToDo: how to lock? as /dev/shm-file?
 ext2 did crash spontaniosly, so I did the memory test
   (bad) mem is used as dsk-cache, good for speed, bad for data integrity
   ToDo: fuse-project crc-fs for ramdisks (tiny-core)
 addr changing from time to time
 memtest: DDR2-333 2GB 2GB/s
   error: 00023e8d023 - 574.5 MB  2015-09-09 kernel badram/badmem-fix?
  locking survives s2ram + s2dsk = suspend-to-disk = hibernate (2 tests)
 kautz18: edac-ce+ue-errors see r_kautz.txt
   pathcc -O2 -fopenmp -D_OPENMP -static -o mem_badram mem_badram.c
   x=$(date +%F_%T);srun -N1-980 -l -W 0 -p big --time=0-9 --cpus-per-task=6 \
      --task-mem=3000 ./mem_badram 2400 -l100;echo $x;date +%F_%T # -l2=2min
 18.03.18 So 07:40 health + srun -N900-970 -l -W0 ... 
  edac-err scx-m10n8  0 0 0 0 0 0 0 0 1 0 0 0           # cpu0/ce
  edac-err scx-m19n11 0 0 0 0 0 0 0 0 481637 0 105584 0 # cpu0/ce dma0/ce
  edac-err scx-m34n2  0 0 0 0 0 0 0 0 24698 0 0 0
  edac-err scx-m34n4  0 0 0 0 0 0 0 0 4 0 0 0
  edac-err scx-m34n7  0 0 0 0 0 0 0 0 3671 0 0 0
  edac-err scx-m34n15 0 0 0 0 0 0 0 0 39478 0 126049 0
  +m34n0 +m34n12
  m34n15: OMP_NUM_THREADS=6 ./mem_badram 2400 -l2;date +%F_%T # 2min
  m19n11: OMP_NUM_THREADS=1 ./mem_badram 2400 -l2;date +%F_%T # 
  m34n12: 
 using OpenMP OMP_NUM_THREADS= 6
edac    0=scx-m34n12.scsystem
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :      6      +6
 malloc 2400 MiB (Mi=2^20) long= 8 B = 64 bit (arg1)
 malloc 1024 blocks to 2400 kiB = 307200 * long (ki=2^10)
 starting, only every 128th block shown ...
 a=0000005569570010 = 341GB blk    0 sz=2400kB (alloc+ini)
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     15      +9
 a=0000005569570010 iLoop=00/16 ffffffffffffffff
ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=0
ERR bad data   a[00008e86]= ffffedffffffffff diff=0000120000000000 j=0 1
ERR blk= 539 range=00008e86 sz=00000001 cnt=1 1
edac    0=scx-m34n12.scsystem
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     36     +21
 loop=  0 speed(w+r):      12806 MB/s (2400.000 MB)
...
# 2nd run 6threads # other address?
 a=00000055697b0010 = 341GB blk    0 sz=2400kB (alloc+ini)
  ...
ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=0
ERR bad data   a[00014e86]= ffffedffffffffff diff=0000120000000000 j=0 1
ERR blk= 539 range=00014e86 sz=00000001 cnt=1 1
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     49      +4
 loop=  0 speed(w+r):      12821 MB/s (2400.000 MB)
 ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=32
 ERR bad data   a[00014e86]= ffffedffffffffff diff=0000120000000000 j=32 1
 ERR blk= 539 range=00014e86 sz=00000001 cnt=1 2
 loop=  0 speed(r):         1836 MB/s (2400.000 MB)
 loop=  0 speed(rh):        2851 MB/s (2400.000 KB)
 ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=32
 ERR bad data   a[00014e86]= ffffedffffffffff diff=0000120000000000 j=32 1
 ERR blk= 539 range=00014e86 sz=00000001 cnt=1 3
 a=00000055697b0010 iLoop=00/16 9bc3679699ccb78c
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     58      +9
 a=00000055697b0010 iLoop=02/16 9bc3679699ccb78c
 a=00000055697b0010 iLoop=04/16 9bc3679699ccb78c
 a=00000055697b0010 iLoop=08/16 9bc3679699ccb78c
 loop=  1 speed(w+r):        809 MB/s (2400.000 MB)
 loop=  1 speed(r):         1862 MB/s (2400.000 MB)
 loop=  1 speed(rh):        2849 MB/s (2400.000 KB)
1 bad blocks detected at task0=scx-m34n12.scsystem, start sleeping ...
for x in $(find /sys/devices/system/edac -name \*e_count);do echo $x $(cat $x);done
/sys/devices/system/edac/cpu/cpu0/L2/ue_count 25
/sys/devices/system/edac/cpu/cpu0/L2/ce_count 58
/sys/devices/system/edac/cpu/cpu0/ue_count 25
/sys/devices/system/edac/cpu/cpu0/ce_count 58   # ue!!! but RAM, not L2
# vs=401MB rss=6MB pid=6374 free=3.4GB
cat /proc/6374/smaps 
55b97d0000-55b9a30000 rw-p 55b97d0000 00:00 0 
Size:              2432 kB
Rss:               2432 kB
Private_Dirty:     2432 kB
cat /proc/6374/maps
55b97d0000-55b9a30000 rw-p 55b97d0000 00:00 0 
dmesg: EDAC DEVICE1: UE: cpu instance: cpu0 block: L2 'L2 Data ECC Error: Address=0x93637bc0'
# addresses changing max_ca 0x934f70c0 1<<31=2GB, 0x13a7c0=2^20=1MB
m34n12 during sleeping 2nd run
m34n12: x=$(date); ./mem_badram 2400 -l2;echo $x;date # -l2=2min
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     58
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ue_count   :     25

2020-01 bb.urz 2 DIMMs DDR 400MHz ECC mc0.csrow6 ce=4000/d+ue=145/225d
   errors on every loop, every bitmap (bios.dimm.interleaved?)
2020-08-10 after some ue on bb.urz mem_badram 
 cat /sys/devices/system/edac/mc/mc0/csrow*?/[cu]e_count|xargs
/sys/devices/system/edac/mc/mc0/csrow0/ce_count 0
/sys/devices/system/edac/mc/mc0/csrow0/ue_count 0
/sys/devices/system/edac/mc/mc0/csrow6/ce_count 1410387
/sys/devices/system/edac/mc/mc0/csrow6/ue_count 300
2020-01-26_10:34 0 0 588076 149
2020-07-30_04:39 0 0 1112702 149
2020-07-30_05:39 0 0 1113236 150
2020-08-04_07:39 0 0 1153225 150
2020-08-04_08:39 0 0 1153442 151
2020-08-08_12:39 0 0 1209076 151
2020-08-08_13:39 0 0 1210331 152  # maxTemp=40C TODO bis 10.08. = abkuehl?
2020-08-10_20:39 0 0 1287115 293 mem_badram lstart=2020-08-10_20:44:29
2020-08-10_21:39 0 0 1287905 295
2020-08-11_09:39 0 0 1297946 298
2020-08-11_10:39 0 0 1298706 300
2020-08-11_10:39 0 0 1298706 300
2020-08-27_13:39 0 0 1410333 300
 # 0-500ue/h peak=2300ue/h L4680-4750 2020-08-07_05:39 - 2020-08-11_09:39
 # maxTemp 2020-08-01_16:48 2020-08-05_20:12 2020-08-06_20:45=39C
 1h logging edac_mc0_row06_ce_ue.log 165kB up=440d ce=30-400/h ue=0
Message from syslogd@ at Mon Aug 10 20:44:30 2020 ...
bb kernel: EDAC MC0: UE page 0x62d9b, offset 0x0, grain 4096, row 6, labels ":": e752x UE from Read
Message from syslogd@ at Mon Aug 10 20:44:32 2020 ...
bb kernel: EDAC MC0: UE page 0x6bb87, offset 0x0, grain 4096, row 6, labels ":": e752x UE from Read
Aug10,2020  00:55s -l2 1800 # rss=855MB=41%*2GB 23MBswap
...  # TODO wrong32bit-output?
ERR bad data   a[0001ffff]= cd84dc39cd84dc39 diff=ffffffffffffffff j=32 131072 r
EERR blk= 839 adr=  0x2b294c0fc010 bad data on thread 0  bad_nb=257
ERR bad data   a[00000000]= cd84dc39cd84dc39 diff=ffffffffffffffff j=32 1 r
...
ERR bad data   a[0001ffff]= cd84dc39cd84dc39 diff=ffffffffffffffff j=32 131072 r
edac    0 i=  -1.00 /sys/devices/system/edac/mc/mc0/ce_count      1287286  +6
ERR edac_blk_bitmap (more than one bad block?)
edac    0 i=  -1.00 rc=6 inv=1 bitmap=0080/0400  blk_idx=00ff   72
... good blocks freed (474 bad blocks left on proc0)
474 bad blocks detected at task0=, sleep + fgets ...
 waiting for input (enter=sleep, char+ret=recheck_bad_blk):



---- hpc21 --- (2x) Intel Xeon Gold 6326 - 2.90GHz, 
 2022-06hpc21 cn155 has sometimes often errors 64B not updated
      lawinenartig?
      rarely this node throws kernel panic
ERR blk= 297 adr=  0x7ef882a83010 bad data on thread 0  bad_nb=2
ERR blk= 297 bad data lrnd= 7545e1467545e146  sum=ffffffffffffffff j=15
ERR bad data   a[00195316]= 8aba1eb98aba1eb9 diff=ffffffffffffffff j=15 1
...
ERR bad data   a[0019531d]= 8aba1eb98aba1eb9 diff=ffffffffffffffff j=15 8
ERR blk= 297 range=00195316 sz=00000008 cnt=8 24
ERR reread to see permanence (caching 30000KB/cores?)
ERR bad data   a[00195316]= 8aba1eb98aba1eb9 diff=ffffffffffffffff j=15 1 r
# diff is ...fff because old value is invers of last value
...
134 bad blocks detected # 2nd run 240GB 32c  17loops
32 bad blocks detected  # 3th run 240GB 32c   17loops
200 bad blocks detected # 4th run 120GB 0-15c 17loops 2*750+220W 35,31,38C
0 bad (2x)              # 6th run 120GB 0-15c  8loops 2*780+170W 35,31,38C
-- no error detected    # 5th run 120GB 16-31c 8loops 2*820+0W   35,31,38C
*/      
//  https://en.wikipedia.org/wiki/ECC_memory
//    spacecraft Cassini–Huygens 1997 2.5Gbit-ECC 280-single-bit-errors/day
//               on day 1997-11-06 4xErrorRate (solar-event)
//    pubs 2007-2009 1e-10..1e-17 errors/bit*h = ca. 1err/GB*h or less
//    google: 8% od DIMMs per year, 7e-11err/bit*h or 5err/8GB*h Streuung?
// mem-intense apps may be more efficient on low-clock (or add cpu-intense app, pe. thermo+psu-stress)
// ToDo: add volatile array, correct speed (less cache dependency)
//  gcc492 -O2 -fopenmp -fverbose-asm -Wa,-adhln -g mem_badram.c
//   ASM? gcc492: unroll8*xmm0(128bit).store + unroll8*rax.cmp
// ToDo: does suspend to disk, reorders phys memory? locking lost?
// ToDo: if found, test time to flip or row-hammer dependence (idle)
//
// bad-mem-successed on:
//  yyyy-mm-dd machine remarks
//  2020-07 t100-2015 trigger bad ECC-DIMM in non-interleaved-channels mode


#include <unistd.h>  /* sleep(60) needs extra h-file */
#include <stdlib.h>  /* system("sleep 600"); */
#include <stdio.h>
#include <sys/time.h>
#include <string.h>   /* strncmp() + memset */
#include <wchar.h>   /* + wmemset */
#ifdef _OPENMP
#include <omp.h>
#endif
#include <sys/mman.h>   /* Memory locking functions */

typedef unsigned long int ULong;

double fstime(){
 struct timeval tv; if(gettimeofday(&tv, NULL))perror("fstime");
 return tv.tv_sec+1e-6*tv.tv_usec; }
 
#define INVLOOPs 16  /* 2015-08, if 1 gcc492 optimizes loops away ToDo: ??? */
#define NB 1024 /* 2020-07 number of blocks 2^n for better unroll, but outer loop */

char * dummy_nodename = "?";
char * slurmd_nodename = "??"; /* SLURMD_NODENAME */
int edac_errors = 0;   /* ... since program start (not the old once) */
/* bit0= invers, bit1= .x.x.x.x, bit2=..xx..xx, bit3=....xxxx bit10=last?  */
/* for 1024 blocks we need 10bit max+1=0x0400 */
int edac_blk_bitmap = 0; // 1<<n for    .x.x.x ..xx..xx ....xxxx  0=init
int edac_blk_invers = 0; // 0 or ~0 for x.x.x. xx..xx.. xxxx....   
int edac_blk_idx = 0 ; // errors triggered by bitmap filter    bitmap>>1
#define NUM_EDAC (8*4+3)  // 35   max. 31 (32bit)
/*
 for x in $(find /sys/devices/system/edac/ -name \*panic_on_ue);do\
  y=$(cat $x);if [ $y != 0 ];then echo $x $(cat $x);fi;done
 for x in $(find /sys/devices/system/edac/ -name \*e_count);do\
  y=$(cat $x);if [ $y != 0 ];then echo $x $(cat $x);fi;done
 # panic_on_ue kautz=yes        t100=no
 # dma0        kautz=yes(ram?)  t100=no
 # cpu0        kautz=RAM_errors t100=no
 # mc0 mc1     kautz=allways0   t100=yes
 # /sys/module/edac_core/parameters/edac_mc_panic_on_ue  kautz=0 t100=0
 # /sys/module/edac_core/parameters/edac_pci_panic_on_pe kautz=0 t100=0
 # /sys/kernel/debug/mce/fake_panic                     kautz=no t100=0
 #  2020-03-26 1t00 trotzdem paniced = reset + B1.double-bit-error displayed
 # 
*/  
char *edacs[NUM_EDAC]={  // better dynnamic?
    "/sys/devices/system/edac/mc/mc0/ce_count",
    "/sys/devices/system/edac/mc/mc0/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc0/ue_count",
    "/sys/devices/system/edac/mc/mc0/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc1/ce_count",
    "/sys/devices/system/edac/mc/mc1/ce_noinfo_count", /* 39 chars */
    "/sys/devices/system/edac/mc/mc1/ue_count",
    "/sys/devices/system/edac/mc/mc1/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc2/ce_count",     // comp2 mc0..mc3 180414
    "/sys/devices/system/edac/mc/mc2/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc2/ue_count",
    "/sys/devices/system/edac/mc/mc2/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc3/ce_count",
    "/sys/devices/system/edac/mc/mc3/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc3/ue_count",
    "/sys/devices/system/edac/mc/mc3/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc4/ce_count",  // epyc18 4chips/socket
    "/sys/devices/system/edac/mc/mc4/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc4/ue_count",
    "/sys/devices/system/edac/mc/mc4/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc5/ce_count",
    "/sys/devices/system/edac/mc/mc5/ce_noinfo_count", /* 39 chars */
    "/sys/devices/system/edac/mc/mc5/ue_count",
    "/sys/devices/system/edac/mc/mc5/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc6/ce_count",    
    "/sys/devices/system/edac/mc/mc6/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc6/ue_count",
    "/sys/devices/system/edac/mc/mc6/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc7/ce_count",
    "/sys/devices/system/edac/mc/mc7/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc7/ue_count",
    "/sys/devices/system/edac/mc/mc7/ue_noinfo_count",
    "/sys/devices/system/edac/cpu/cpu0/ce_count",  // kautz
    "/sys/devices/system/edac/cpu/cpu0/ue_count",  // kautz panic_on_ue==0?
    "/sys/devices/system/edac/dma/dma0/ce_count"};
// /sys/devices/system/node/node[01]/numastat t100.TODO numa_hit

long int chk_edacs=(1l<<NUM_EDAC)-1; /* bit pattern? bit/path_in_list_above? */

// analyse memory given by pointer and length in sizeoflong
//  gcc492 -O2 -funroll-loops ... ToDo: optimiert loops to cache?
//    num = mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
//    v2015-08: *a=512MB to 512kB-Blocks/core (L2-cache) = Bad (ToDo **a)
//  ToDo: multiple copy mini-array(4kB)
// use volatile to avoid bad optimization, we expect possible memory change
//   problem volatile versus vectorization? check asm-code!
//   gcc492 -O2 -fopenmp -fverbose-asm -Wa,-adhln -g mem_badram.c | less # ASM-code
// all blocks in INV-loop to avoid in-cache work
// old 2017-- write all (200GB/100s), read all (200GB/100s)
//      100s active measure time between write and next read,
//      but +100s useless time between read and next write
// ToDo: better write new pattern to block after read+check a block
//      1/1000 useless time + double measure time (option: initial = skip check)
//      also writing in neighbourhood of reading is kind of hammering?
//      measure time expandable by SIGSTOP

// added 2018-03 to get ECC response, j=0..INVLOOP-1
int chk_edac(int procid, int ii, int j, ULong *bad, int *bad_nb) { // j=LOOP.j
  FILE *p1=NULL;
  // added 2017-08 Todo: skipmodus on err
  int i1, rc=0, i2[NUM_EDAC]={0,0,0,0,0,0,0,0,0,0,0};
  static int    i3[NUM_EDAC]={0,0,0,0,0,0,0,0,0,0,0};
  static int out_nodename=1; /* output once only on first error */
  // on error reduce NUM_EDAC by removing mc4-mc7 entries
  if (NUM_EDAC > 8*sizeof(chk_edacs)) exit(5);
  if (chk_edacs) { // set if there are /sys/.../ce_counts
    // if (edac_errors)
    usleep(100*1000); // added 100ms 2018-04-04, edac seems to need relax time
    // without that time binary tree bad block search does work badly 
    //  ... finds different blocks, maybe because edac delay
    for (i1=0;i1<NUM_EDAC;i1++)
    if (1 & (chk_edacs >> i1)) { // loop over path list 
      p1=fopen(edacs[i1],"r");
      if (!p1) { // suppress 2nd failed-message 2020-06
        if (procid==0 &&  i1>=1 && 1==(1&((chk_edacs<<1)>>i1)) )
          printf("chk_edac.open %-47s ok ...\n",edacs[i1-1]);
        if (procid==0 && (i1==0 || 1==(1&((chk_edacs<<1)>>i1))))
          printf("chk_edac.open %-47s failed ...\n",edacs[i1]);
        chk_edacs &= ~(1l<<i1); } // added 2018-03 for non-ECC
      if (p1 && fscanf(p1,"%d",&i2[i1])!=1)
        printf("chk_edac.read %-47s failed ...\n",edacs[i1]);
      if (p1 && fclose(p1)) perror("err fclose");
      if (i2[i1]!=i3[i1]) {
        printf("edac %4d i=%4d.%02d %-45s %5d %+3d\n",
           procid, ii, j, edacs[i1], i2[i1], i2[i1]-i3[i1]);
        rc += (i2[i1]-i3[i1]); // return new errors
        i3[i1]=i2[i1];
      }
      // if (i1>1 && chk_edacs==0) break;
    }
  }  // chk_edacs,
  //  ### popen gives no err if cat failes
  //p1=popen("cat /sys/devices/system/edac/mc/mc[01]/[cu]e_{,noinfo_}count"
  //   " /sys/devices/system/edac/cpu/cpu[0-5]/[cu]e_count|xargs 2>&1","r");
  // for (rc=i1=0;i1<NUM_EDAC;i1++) rc += i2[i1];
  if (rc && out_nodename) { out_nodename=0; /* output hostname once only */
    printf("edac %4d %s rc=%d\n", procid, slurmd_nodename, rc);
  }
  // if no error switch blk_inverse, on error set bit, next bit if bit found
  if (rc && edac_blk_bitmap) {
    // 2018-04-04 add initial blk_idx for new cycle
    //   ToDo: after 2 detects, mark blk as bad and go around!?
    if (edac_blk_bitmap==1) { edac_blk_idx=0; }
    // bitmap & 1 == 0 for .x.x.x, == 1 for x.x.x.
    if (!edac_blk_invers) edac_blk_idx |= edac_blk_bitmap;
    else if (edac_blk_idx & edac_blk_bitmap)
      printf("ERR edac_blk_bitmap (more than one bad block?)\n");
    printf("edac %4d i=%4d.%02d rc=%d inv=%d bitmap=%04x/0400"
           "  blk_idx=%04x %4d\n",
          procid, ii, /* invloop */ j, /* ecc-errors */ rc, 
          edac_blk_invers&1, edac_blk_bitmap,
          edac_blk_idx, edac_errors);
    if (edac_blk_bitmap==1<<9) { /* final of 1024 blocks */
      static int last_blk=0, repeat=0;
      if (last_blk==edac_blk_idx) { repeat++;
        if (repeat>2) { bad[last_blk]++; bad_nb++; // ToDo better recheck maskbits?
          printf("WARNING: %3d mark block %4d 0x%04x as bad, continue ...\n",
                 procid, last_blk, last_blk); repeat=0; last_blk=0;
        } /* mark blk as bad + continue (further blocks?) */
      } else
      if (repeat==0) { last_blk=edac_blk_idx; repeat++; } else
      { /* block changing */ }
      printf("edac %4d i=%4d.%02d bad block identified, blk_idx=%04x\n",
          procid, ii, /* invloop */ j, edac_blk_idx);
      // bad[edac_blk_idx]++; TODO, see above
    } // edac_blk_bitmap==1<<9
  } // if (rc && edac_blk_bitmap)
  return(rc);
} // end edac-pipe-module chk_edac()
  
#ifndef VSIZE
#define VSIZE 16  // vectorization loop inside main-loop Jan22
#endif

// TODO because read cache-line before write, a read-modify-write is faster
//   if memory counted twice (rd+wr)!
int mem_wr(ULong volatile ** pa,
           const long nb, const long num, const ULong rnd,
           int procid, ULong *bad){
  long i, j, k;  ULong lrnd, vrnd[VSIZE];  ULong volatile * a;//, *b;
// TODO default(none)
// TODO howto use lower half threads and let upper half (vthreads) do CPU-stress 
//      or start a parallel nice -19 cpu-stress on vthreads?
#pragma omp parallel for schedule(static) default(shared) private(i,j,k,a,lrnd,vrnd)
  for (i=0;i<NB;i++) { /* 1024 blocks writing */
    a = pa[i]; lrnd=rnd; if (bad[i] || !pa[i]) continue; // 2018-03 skip bad blks
    // bitmap & 1 == 0 for .x.x.x, == 1 for x.x.x.
    if (((i^edac_blk_invers)&edac_blk_bitmap) || !edac_blk_bitmap) {
#if 0 // rep stosq (64bit data)
      wmemset((void *)a,(wchar_t)lrnd,2*num);// write through caches?
#elif 0 && defined(__x86_64__) // newer cpus can do rep stos at max speed
  // v2df vector of 2 double-floats? 128b?, v2si=2*32b v4hi=4*16b
  // this would not help if we want to write different values
  //  but rep generates faster micro code inside modern cpus
  /* cld: clear direction flag == increment pointer on rep (std decr.) */
  /* rep stosq: repeat until decrement ecx is zero or spez. condition */
  // asm [volatile] ("asm" : out-op-list : in-op-list [: modify-list]);
  //   volatile means in-ops will be modified (?) 
  // fails on t100+gcc-8.2, but t100+gcc-4.8.w=50%read
  //  att(AT&T)-syntax(movl $1,%eax) differs from intel(mov eax,1)
  //  a=eax,ax,al  c=ecx,cx,cl  S=esi,si D=edi,di
// gcc4.3 adds: (missing on gcc8.2)
//        movq    %r12, %rcx
//        movq    %rbp, %rax
// default: -masm=att (or set -masm=intel)
    k=7;
    asm volatile ("cld\n"
      "rep stosq"
      : "=D" (b), "=c" (k) /* no output registers or dummys+volatile gcc8 */
      : "D" (a) /*dest*/, "c" (num) /*count*/, "a" (lrnd) /*val*/
      :  "cc" /*, "memory"*/ ); //  cc=flags + memory or bx, ebx, %ebx
      // Reihenfolge ist egal, Namen sind x86 spez. c=%ecx a=%acx
      // stosq=store-string-4B ToDo scasq=scan_string cmpsq=compare-string
      //  repne/repe scas not micro-code-optimized
      if (k!=0) printf ("cld stosq %%rcx=%ld\n", k); /* expect 0 */
#else
      for (j=0;j<VSIZE;j++) vrnd[j]=lrnd; // ini vector-rnd[]
      //for (k=0; k<num; k++) a[k]=lrnd; // 2015-08-31 rnd changed to lrnd
#pragma GCC unroll 1  // TODO better use fix size (pagesize?)
      for (k=0; k<(num/64)*64; k+=VSIZE) { // tell compiler loop is %64
#pragma GCC unroll 16
        for (j=0;j<VSIZE;j++) { a[k+j]=vrnd[j]; } } // Jan22 vectorize
#endif
    } // if good+valid memory block to test (bin-tree on ecc bad block search)
  } // for i<NB block-loop (big data avoids cache intern work) 2015-09-03 2nb-loops
  return 0;
} // mem_wr

#include <stddef.h>  // _mm_prefetch?
#ifdef __SSE4_1__
#include <smmintrin.h>
#endif

#ifdef __AVX__
#include <immintrin.h>
#endif
#include <emmintrin.h> // hpc18+gcc8

// t100.n146 86-93GB/s=89.5+-4%, why not const?  2019-01
int mem_rd(ULong volatile ** pa,
    const long nb, const long num, const ULong rnd,
    const int procid, ULong *bad, int rj){ // rj = round index
  long j1=0, j2=0, /* addr range */ jn=0, /* num */ i, j, k,
     sum=0, vsum[VSIZE];
  ULong lrnd = rnd, vrnd[VSIZE];
  int bad_nb=0; /* bad blks counter (bad_NumBlocks) */
  static long jn2=0;
  ULong volatile * a;
  // Jan22 less bus with default(none) instead shared
  // Jun22 add rnd,procid,num to firstprivate gcc-11 generates errors else
#pragma omp parallel for schedule(static) default(none) shared(rj) \
  private(a,j,k,j1,j2,jn,sum,lrnd,vrnd,vsum) \
  firstprivate(rnd, procid, num, bad, edac_blk_invers, edac_blk_bitmap, slurmd_nodename,pa)\
  reduction(+:jn2,bad_nb)
  for (i=0;i<NB;i++) { /* 1024 blocks reading */
    a = pa[i]; lrnd=rnd; for(j=0;j<VSIZE;j++) vrnd[j]=lrnd;
    // 2018-04-03 try to fix bad data ffffff after EDAC error
// ???    if (pa[i]) if (pa[i][0]==~rnd) lrnd=~rnd; // 2018-04-03 bad fix
    if (bad[i] || !pa[i]) continue; // 2018-03 skip bad blks, find others too 
    if (((i^edac_blk_invers)&edac_blk_bitmap) || !edac_blk_bitmap) {
      // ToDo: vectorize lrnd
      sum=0; j1=j2=jn=0; for(j=0;j<VSIZE;j++) vsum[j]=0; // Jan22 +vsum
// TODO parallelize outer loop i=0..1023  asm L47 unroll8*xorq no-vec gcc8
//   no speedup with sum1+2 lrnd1+2 k+=2 ... k+=4
//  TODO 2modes: parallel k or parallel i (test speed non-interleaved!)
//asm("repne scasq"
//    : "=c" (count), "+D" (p)
//    : "m" (*(const char (*)[]) p), "c" ((size_t)num), "a" (lrnd));
//  until 2020-06 the inner loop was parallel (slow on non-interleaved mem)
//#pragma GCC ivdep  // vectorize loop (see xorshift8M-loop) 
//#pragma omp parallel for default(shared) private(k) reduction(|:sum)
//    (num/64)*64 let the compiler now, that the loop is 64-multiple
#pragma GCC unroll 1  // Jan22 2=-0.5% 4=-1%
      for (k=0; k<(num/64)*64; k+=VSIZE) {   // Jan22 +VSIZE
        // TODO use BSIZE=8*4kB=8*512*8B instead of num=240MB/1024/8B=30k
        //   but exclude low-percent bad-mem
        // by experimenting speedup on prefetch k+128 (128*8=1024B L1 cache)
        //  prefetch ca. 100-200 cycles before need
//    T0 (temporal data)—prefetch data into all levels of cache
//    T1 (temporal data with respect to L1 misses)—prefetch into L2 and higher.
//    T2 (temporal data with respect to L2 misses)—prefetch into L3 cache and higher, or an implementation-specific choice.
//    NTA (non-temporal data with respect to all cache levels)—
//        prefetch data into non-temporal cache structure
//        and into a location close to the processor, minimizing cache pollution.
        _mm_prefetch((const void *)&a[k+256  ], _MM_HINT_NTA);
#if VSIZE > 8
        _mm_prefetch((const void *)&a[k+256+8], _MM_HINT_NTA); // cline=8*8B
#endif
#pragma GCC unroll 16
        for (j=0;j<VSIZE;j++) { // Jan22
          vsum[j] |= a[k+j] ^ vrnd[j]; } // 2015-09-03 vectorize +5%
      } // longest k-loop
      sum=0;for(j=0;j<VSIZE;j++) sum |= vsum[j]; // Jan22 reduce vector
      // 2022-06 hpc21.cn206 1x su-error (high32bit) but mem=ok
      /* 2nd slow scan on error! possible it reads from CPU cache now? */
      if (sum) { // bit error handling (optimized away? use volatile array?)
        if (bad_nb<64 || !(bad_nb & (bad_nb-1))) // use bad_blk_count
          printf("ERR blk=%4lu adr=%16p bad data on thread %d %s bad_nb=%d\n",
                          i, a, procid, slurmd_nodename, bad_nb+1);
        if (jn2==1024) printf("ERR suppress further messages ...\n"); // 2020
        if (jn2<1024 || (jn2 & (jn2-1))==0)
          printf("ERR blk=%4lu bad data lrnd= %016lx  sum=%016lx j=%u\n",
            i, lrnd, sum, rj /* round */);
        /* allows ^c on long scrolls */
        if (jn2>1023 && (jn2 & (jn2-1))==0) usleep(100000); /* 100ms */
        j1=j2=jn=0;
        for (k=0; k<num; k++) if (a[k]!=lrnd) { // slow reread
          // critical global vars
          jn++; jn2++; j2=k+1; /* j2>j1 means bad range set */
          if ( (jn2<1024 || (jn2 & (jn2-1))==0)
            && (jn<16 ||    (jn  & (jn-1))==0) )
          printf("ERR bad data   a[%08lx]= %016lx diff=%016lx j=%u %lu\n",
            k, (ULong)a[k], (a[k]^lrnd), rj /* round */, jn);
            /* we may check for last lrnd here (lost writes 2022.cn155) */
        } else { if (j1==j2) j1=j2=k+1; }
        if (j2==j1) printf("ERR blk=%4lu bad data could not localized\n",i);
        else
          if (jn2<1024 || (jn2 & (jn2-1))==0)
             printf("ERR blk=%4lu range=%08lx sz=%08lx cnt=%lu %lu\n",
                     i, j1, j2-j1, jn, jn2);
        if (jn2<1024 || (jn2 & (jn2-1))==0)
          printf("ERR reread to see permanence (caching %ldKB/cores?)\n",
            num/1024l);
        /* kautz 3GB/6c/1000=500kB vs. 256KB-L2
         * t100  250GB/16c/1000=15.6MB vs. 256KB-L2 */
        j1=j2=jn=0;
        for (k=0; k<num; k++) if (a[k]!=lrnd) { 
          jn++; j2=k+1; /* j2>j1 means bad range set */
          if ((jn2<1024 && jn<16) || (jn & (jn-1))==0) /* "r"=reread */
          printf("ERR bad data   a[%08lx]= %016lx diff=%016lx j=%u %lu r\n",
            k, (ULong)a[k], (a[k]^lrnd), rj /* round */, jn);
        } else { if (j1==j2) j1=j2=k+1; }
        if (j2==j1) printf("ERR blk=%4lu bad data could not localized\n",i);
        else
          if (jn2<1024 || (jn2 & (jn2-1))==0)
             printf("ERR blk=%4lu range=%08lx sz=%08lx cnt=%lu %lu reread\n",
                     i, j1, j2-j1, jn, jn2);
      }
      if (jn/* || sum*/) { bad[i]++; bad_nb++; jn=0; } // count bad blocks + mark as bad
    } // if good block (no errors until yet) or half blocks on ecc-search
  }  // 1024 mem-blocks i<NB  omp parallel !!! reduction+: bad_nb,jn2
  return bad_nb;
} // mem_rd


// write + readchk (better read and check old + write for next?)
int chkmem(ULong volatile ** pa,  /* array of mem blocks, return: num errors */
    const long nb, long num, /* 2D-array params nb=1024(or less on failure) */
    int ii, /* loop count 0..99 for compact debug output */
    ULong rnd, /* random 64bit-number */
    int procid, /* slurm_proc_id=0..n-1 */
    ULong *bad,  /* marker array [1024] */
    int sum_bad_nb) /* for dbg num_bad_blocks state, better outside chkmem? */
{
  double t1,t2,t3,t4; /* timing vars */
  static double tw_max=0,tw_min=0,tr_max=0,tr_min=0; // 2020-06 timing vars
  ULong lrnd=rnd /* rnd-seed */; int i, j /* round-idx */, rc_bad_nb=0;
  int bad_nb=0, i_bad0=0, i_bad1=0; /* outputting from first_bad_blk to last_bad_blk */
  // char s2[1024]={0}; // buffer edac logs (see below)
  for (j=0;j<INVLOOPs; j++){ /* fix20150902: why j=1 ... ? bad measure! */
    lrnd=~lrnd;  /* inverse bitpattern between tests */
    /* j=0..7 ~rnd rnd ~rnd rnd ... ~rnd rnd */
    /* to see state of errors on screen at first sight even after 100 loops */
    if (procid==0 && ((j&(j-1))==0) /* 180414 log2 */
                  && ((ii&(ii-1))==0 || ii<8)){ /* log reduced output */
      bad_nb=0;
      for (i=0;i<NB;i++) { if (bad[i]) break; } if (i<NB) i_bad1=i_bad0=i;
      for (   ;i<NB;i++) if (bad[i]) { i_bad1=i; bad_nb++; }
      //printf(" a=%016lx iLoop=%4d.%02d/%d %016lx bad= %4d %4d nb %4d %4d\n",
      //  (ULong)pa[0], ii, j, INVLOOPs, lrnd,
      //   i_bad0, i_bad1 /* blk range */, bad_nb, sum_bad_nb+rc_bad_nb);
      printf(" iLoop=%4d.%02d/%d %016lx bad= %4d %4d nb %4d %4d\n",
        ii, j, INVLOOPs, lrnd,
        i_bad0, i_bad1 /* blk_range */, bad_nb, sum_bad_nb+rc_bad_nb);
/* ToDo: add output which block-range? show only log2(ii) or 0..3 + runtime */
      fflush(stdout); // slow?
      /* slurm needs flush to see progress 2015-09-03 */
    }
#pragma omp barrier // sonst overwrite end block i, start blk i+1 ???
    t3=fstime();
                mem_wr(pa, nb, num, lrnd, procid, bad);   // WRITE MEM
    t4=fstime();
#pragma omp barrier
    // ToDo additional variing sleep here?
    { int ce0;
      usleep(100*1000); // 100ms to get all relevant interrupts or polls
      ce0= chk_edac(procid,ii,j,bad,&rc_bad_nb); // sleep 1s before check
      if (ce0) printf("Warning: EDAC errors during write (ce=%d)\n", ce0);
    }
#pragma omp barrier
    t1=fstime();
    rc_bad_nb+= mem_rd(pa, nb, num, lrnd, procid, bad, j); // READ MEM
    t2=fstime();
    // TODO half blocks on ecc-error search (edac_blk_bitmap)
    if (procid==0 && t1!=t2 && t4!=t3 /*&& (ii&(ii-1))==0 && (j&(j-1))==0*/){
      int fr1=0,fw1=0; /* output flag */
      if (tw_max==0.0 || tw_max<t4-t3){ tw_max=t4-t3;fw1=1;} // 2022-06 reduce output
      if (tw_min==0.0 || tw_min>t4-t3){ tw_min=t4-t3;fw1=1;} //  ... + sum range
      if (tr_max==0.0 || tr_max<t2-t1){ tr_max=t2-t1;fr1=1;}
      if (tr_min==0.0 || tr_min>t2-t1){ tr_min=t2-t1;fr1=1;}
      if (fw1) // print-out flag
        printf(" iLoop= %3d.%02d speed.w: %10.3f - %10.3f GiB/s (%.0f MiB)\n",
             (int)ii, (int)j, ((edac_blk_bitmap)?0.5:1.0)*
              1.*nb/1024.*num*sizeof(pa[0])/1024./1024./tw_max, // = mb/t
              1.*nb/1024.*num*sizeof(pa[0])/1024./1024./tw_min, // = mb/t
                          num*sizeof(pa[0])/1024.*1.*nb/1024.);
      if (fr1)
        printf(" iLoop= %3d.%02d speed.r: %10.3f - %10.3f GiB/s (%.0f MiB)\n",
             (int)ii, (int)j, ((edac_blk_bitmap)?0.5:1.0)*
              1.*nb/1024.*num*sizeof(pa[0])/1024./1024./tr_max, // = mb/t
              1.*nb/1024.*num*sizeof(pa[0])/1024./1024./tr_min, // = mb/t
                          num*sizeof(pa[0])/1024.*1.*nb/1024.);
      if (fr1||fw1) fflush(stdout);
    } // print speed w/r
    // echo 0 > /sys/devices/system/edac/mc/edac_mc_panic_on_ue # dflt=1
    // echo 0 > /sys/devices/system/edac/cpu/panic_on_ue        # dflt=1
    // mkautz: /sys/devices/system/edac/mc/mc[01]/[cu]e_{,noinfo_}count
    //         /sys/devices/system/edac/cpu/cpu[0-5]/[cu]e_count # m34n15
    //         /sys/devices/system/edac/dma/dma0/[cu]e_count     # m34n15
    //         /sys/devices/system/edac/fsw/fsw0/ce_count
    //         /sys/devices/system/edac/dma/panic_on_ue # 1
#pragma omp barrier
    { int ce0;
      usleep(100*1000); // 100ms
      ce0= chk_edac(procid,ii,j,bad,&rc_bad_nb); // sleep 1s before read?
      if (ce0) edac_errors += ce0; // static var
      // change bitmap if we had errors on inv=0 or inv=1
      //  try to find the bad memblock wr/rd .x.x.x x.x.x. ..xx..xx xx..xx
      if (edac_blk_bitmap) edac_blk_invers ^= ~0; // until ce0 found
      if (ce0) {
        if (edac_blk_bitmap==0) { edac_blk_bitmap=1; edac_blk_invers=0; }
        else {
           if (edac_blk_bitmap==1<<9) edac_blk_bitmap=1; else
           if (edac_blk_bitmap)       edac_blk_bitmap<<=1;
           edac_blk_invers=0;
        }
        // ToDo better 16*alternation and counting errors for each invers?
        //   if we have errors in both, its not a specific block
        if (edac_blk_bitmap<=32) // inv alternates until next edac error
          printf("INFO test bitmap=0x%04x/0x0400 inv=%d,1,0,1,...\n",
                edac_blk_bitmap, edac_blk_invers&1);
      } // edac-error found
    }  // chk_edac
    // j&1 leaves memory written with lrnd (not ~lrnd)
    //  difficult to detect case odd loop break and edac_blk_pattern written
    if (rc_bad_nb && (j&1)) break; // bad for rereads if invers written (even)
  } // INV-loop j=0..15
  return rc_bad_nb; /* return code is num bad_nb for lrnd 16 loops */
} // chkmem

#define PAGEMAP_LENGTH 8  // kernel specific, check it!
unsigned long get_page_frame_number_of_address(void *addr) {
   // Open the pagemap file for the current process
   FILE *pagemap = fopen("/proc/self/pagemap", "rb"); // not in 2.6.18 RH5
   // RH5 2.6.18 /proc/self/smaps (txt)
   // The page frame number is in bits 0-54 so read the first 7 bytes and clear the 55th bit
   unsigned long page_frame_number = 0;
   // Seek to the page that the buffer is on
   unsigned long offset = (unsigned long)addr / getpagesize() * PAGEMAP_LENGTH;
   if(pagemap && fseek(pagemap, (unsigned long)offset, SEEK_SET) != 0) {
      fprintf(stderr, "Failed to seek pagemap to proper location\n");
      exit(1);
   }
   if (pagemap) fread(&page_frame_number, 1, PAGEMAP_LENGTH-1, pagemap);
   // page_frame_number &= 0x7FFFFFFFFFFFFF;
   if (pagemap) fclose(pagemap);
   return page_frame_number;
}


int main(int argn, char *argv[]){
  // todo add 1kB for prefetch behind array? 
  static ULong volatile * pa[NB]; /* no stack, segfault at getenv */
  ULong rnd=0, rnd1, rnd2; long ii, i, j;
  static ULong bad[NB]; /* no stack */
  long num, nb=0, mb=1024; // 2015-10 signed for pathcc openmp
  char *sprocid=NULL;
  int n_loops=100; /* number of loops (of INVLOOPs with different RND) */
  int procid=0, num_nodes=1; /* slurm_procid */
  int verbose=0, bad_nb=0; /* 2018-04-04 bad_nb changed from long to int */
//  int verbose=1; /* 0=quiet(for multinode) ToDo: check slurm_env! */
#ifdef _OPENMP
  int omp_n= 1;  /* openmp-threads */
#endif
  int sl=sizeof(ULong);            /* sizeLong should be 8 on 64bit or 4 on 32bit */
  // if (argn>1) mb=atoi(argv[1]);    /* memory in MB */
  for (i=1;i<argn;i++) {
    if (argv[i][0]>'0'  && argv[i][0]<='9') mb=atoi(argv[i]); /* MB */ else
    if (argv[i][0]=='-' && argv[i][1]=='l') n_loops=atoi(argv[i]+2); else
    if (argv[i][0]=='-' && argv[i][1]=='v') verbose=atoi(argv[i]+2); else
     { printf(
    "usage: mem_badram [<megabyte=1000>] [-l<nloops=100>] [-v<verbose=0>]\n");
      exit(1); }
  }
  num= mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
  if (num & 63) {
    printf("WARNING: num%%63=%d , round down\n", (int)num&63);
    num &= ~63l;
  }
  // SLURM_PROCID  #  0..SLURM_NPROCS
  if (verbose>6) printf("DBG L%d\n",__LINE__);
  sprocid=getenv("SLURM_PROCID");  /* string: 0..n or "" 2018-03 segflt */
  if (!sprocid) 
    sprocid=getenv("OMPI_COMM_WORLD_RANK"); // from OPENMPI
  if (sprocid) procid = atoi(sprocid); /* id0 is talking, id1++ is silent */
  slurmd_nodename=getenv("SLURMD_NODENAME");  /* nodename or "" */
  if (verbose>6) printf("DBG L%d host=%s\n",__LINE__,slurmd_nodename);
  if (!slurmd_nodename) // kautz18 geht erst nach export HOSTNAME, set, ~env
    slurmd_nodename=getenv("HOSTNAME");  /* nodename or "" CentOS5 */
  if (verbose>6) printf("DBG L%d host=%s\n",__LINE__,slurmd_nodename);
  // ToDo: /proc/sys/kernel/hostname = unistd.gethostname(char *,size)
  if (!slurmd_nodename)
    { static char s1[128]={0}; gethostname(s1,128); // 2018-03-18
      if (s1[0]) slurmd_nodename=s1; }
  if (!slurmd_nodename)
    slurmd_nodename=dummy_nodename; /* we dont want NULL-pointer "?" */
  if (verbose>6) printf("DBG L%d host=%s\n",__LINE__,slurmd_nodename);
  // 2018-03 atoi(sprocid) = SEGFAULT
  { char    *s1=getenv("SLURM_NPROCS");           /* slurm? */
    if (!s1) s1=getenv("OMPI_COMM_WORLD_SIZE");   /* openmpi? */
    num_nodes=(s1)?atoi(s1):1; }
  if (num_nodes==0) num_nodes=1;
  if (verbose>6) printf("DBG L%d\n",__LINE__);
  if (!procid) {
//SLURM_NNODES=1
//SLURM_NPROCS=1
//SLURM_NODELIST=scx-m0n0
    system("date +%F\\ %T"); // 2018-04-14 for better output-documentation
    printf("mem_badram bad RAM detection 2012-2022.06 by JS %luMB l%d v%d\n",
      mb, n_loops, verbose);
    printf(" long=%dB=%dbit, wchar=%dB=%dbit, size_t=%dbit VSIZE=%d\n",
            sl /* sizeofLong */, sl*8 /* bits */,
            (int)sizeof(wchar_t) , (int)sizeof(wchar_t)*8 /* bits */,
            (int)sizeof(size_t), VSIZE/* int64 vector size */);
    printf(" NUM_PROCS= %d %s\n", num_nodes, getenv("SLURM_NODELIST"));
  }
  if (procid<3) {
    printf(" SLURM_PROCID=%s = %d %s (skip3++)\n",
       sprocid, procid, slurmd_nodename); /* 0..n or "" */
    fflush(stdout);
  }
#if defined(__GNUC__) // 2020-06 for better documentation
  if (!procid) printf("Compiler: GNUC %d.%d.%d\n",
                  __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif
  if (verbose>6) printf("DBG L%d\n",__LINE__);
  if (num_nodes<2) slurmd_nodename=""; // short line on single nodes
#if defined(_OPENMP)
  // 199810=1.0(gcc-4.2,icc10) 200505=2.5(gcc-4.1) _OPENMP=YYYYMM openmp-spec
  // 200805=3.0(gcc-4.4)       201107=3.1(gcc-4.7,4.8)
  // 201307=4.0(gcc-4.9)       201511=4.5
  if(procid==0) printf(" OpenMP-Support: OPENMP=%d\n",_OPENMP);
#pragma omp parallel default(shared)
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  // output can be used by copy and paste
  // OMP_NUM_THREADS=16 GOMP_CPU_AFFINITY="0-31" ./mem_badram 120000
  if(procid==0)
    printf(" OMP_NUM_THREADS=%d GOMP_CPU_AFFINITY=\"%s\" ./mem_badram %lu -l%d\n",
           omp_n, getenv("GOMP_CPU_AFFINITY"), mb, n_loops); // or "(NULL)"
  if(procid==1)
    printf(" OMP_NUM_THREADS=%d GOMP_CPU_AFFINITY=\"%s\" (slurm=1)\n",
           omp_n, getenv("GOMP_CPU_AFFINITY")); // or "(NULL)"
#endif /* _OPENMP defined */
  chk_edac(procid,0,0,bad,&bad_nb); // 1s + print initial edac values, panic_on_ue shoud be 0
  if (!procid) {
    printf(" malloc %lu MiB (Mi=2^20) (arg1)\n", mb /* sumMB */);
    // split into 1024 blocks of size mb*1024
    printf(" malloc %d blocks to %lu kiB = %lu * long (ki=2^10)\n", NB, mb, num);
    printf(" starting, only every 128th block shown ...\n"); fflush(stdout);
  }
  // TODO20 check overcommit memory, should be 2
  fflush(stdout);
  // serial loop malloc
  for (i=0;i<NB;i++) { // 1024 memory blocks * 128MiB = 128GB
    bad[i] = 0; /* marked as clean */
    pa[i] = (ULong *) malloc(mb*1024l); /* alloc mb*kB (memoryMB/1024) */
    // usleep(100000);
    if (procid==0 && ((i&127)==0 || !pa[i]))
      printf(" a=%016lx = %3luGB blk %4lu sz=%4luKiB (alloc+ini)\n", // old: touch
             (ULong)pa[i], ((ULong)pa[i])>>30, i, mb);
    if (!pa[i]) { break; }  // malloc failed
  } nb=i;
  if (nb!=1024) 
    printf(" WARNING: only %lu blocks = %lu kiB allocated on %s\n",
           nb+1, (nb+1)*mb, slurmd_nodename);
  fflush(stdout);
  if (!nb) exit(1);
  // 2020-07 max. speed with touching memory thread local (numactl -l)
  //   + in order for non-interlaced memory
  //   t100-i1:1+static,1 ordered w19-43 r26-66 GB/s numa_bala=0 w20 r25
  //   t100-i1:1+static  +ordered w43    r65-66 GB/s numa_bala=0 w43 r66 
  //   t100-i1:1+static  -ordered w39-40 r60-61 GB/s numa_bala=0 w43 r64
#pragma omp parallel for ordered schedule(static) private(i,j) default(shared)
  for (i=0;i<1024;i++) { // 1024 memory blocks * 128MiB = 128GB
    // touching j+=512 (*8B=4kB) is faster, but gives bad binding on openmp
    // cat /sys/devices/system/node/node[01]/numastat t100.TODO numa_hit
    //  see t100.node064
    //  TODO better parallelize outer loop over i=0..1023
    if (!pa[i]) continue;
//  ordered touch for non-channel-interleaved multi-channel memory
//#pragma omp parallel for default(shared) private(j)
// #pragma omp ordered // this is for 3*speed on 4 non-interleaved memory channels
//                  Jan22 ordered removed, ini speedup, no overall slowdown
    for (j=0; j<num; j++) pa[i][j]=0; // single task faster if j+=512 (8B)
  } // num blocks (serial)
#if 0  // do later if bad block found!? hpc18-speed-problems?
  if ( mlockall(MCL_CURRENT|MCL_FUTURE) == -1 ) // 2020-06 does seqini of unini
    printf("ERROR: mlockall failed (non-root?)\n"); /* root only */
#endif
  for (i=0;i<1024;i++) { // 1024 memory blocks * 128MiB = 128GB
    // ToDo show HW-Adresses
    unsigned long page =
      get_page_frame_number_of_address((void *)pa[i]);
    if (procid==0 && (i&127)==0 && pa[i])
      printf(" a=%016lx = %3luGB blk %4lu sz=%4luMiB pageframe=%016lx\n", // old: touch
             (ULong)pa[i], ((ULong)pa[i])>>30, i, mb/1024, (ULong)page);
  }
  { int ce0=chk_edac(procid,-1,0,bad,&bad_nb);  // 1s + chk edac
    if (ce0) edac_errors+=ce0; }
  for (ii=0;ii<n_loops;ii++) { // dflt 100 rounds
    int rc=0; 
    rnd1=rand(); // random fill pattern, ToDo better as seed for xorshift-fill
    rnd2=rnd1;   // rnd2=rand(); 2020-07-09 for wmemset function
    rnd=(rnd1<<32) | rnd2;  // 64bit random from 32bit (wmemset)
    if (/*ii!=0 &&*/ (ii&7)==0) rnd=0; /* 1803: ii!=0 */
    // printf(" rnd %016lx %016lx %016lx\n", rnd1, rnd2, rnd);
    // return number bad blocks, wr+rd
    bad_nb+=(rc=chkmem(pa,nb,num,ii,rnd,procid,bad,bad_nb)); // 0=OK 16*invert+write+read+cmp
    // 2018-03 do not break loop, just skip bad blocks on further search
    usleep(60000); // 60ms (extra errors?)
    // getconf -a | grep -i cache # L1+L2_CACHE_SIZE+ASSOC=8+LINESIZE=64
    bad_nb+= mem_rd(pa, nb, num, rnd, procid, bad, 32);
    edac_errors += chk_edac(procid,-1,0,bad,&bad_nb);  // 1s + chk edac 2020-06
    // t100.n146 86-93GB/s=89.5+-4%, why not const?  2019-01
    // ToDo: within 65ms DRAM refresh do 130e3 uncached reads
    
    // latency: L1=1ns=light/30.5cm L2=3ns L3=12ns lMem=60ns rMem=100ns
    // ToDo: core1: full-L1-loop core2: access remote RAM or vv 
    mem_rd/*_hammer*/(pa, nb, num, rnd, procid, bad, 32);
    edac_errors += chk_edac(procid,-1,0,bad,&bad_nb);  // 1s + chk edac 2020-06
    bad_nb+= mem_rd(pa, nb, num, rnd, procid, bad, 32);
    edac_errors += chk_edac(procid,-1,0,bad,&bad_nb);  // 1s + chk edac 2020-06
 
    // Todo 2nd read after some seconds (+partly write = rowhammer?)
  } // loop100
  for (bad_nb=i=0;i<nb;i++) // 2018-04-05 add set NULL on free, avoid segflt
    if (pa[i] && !bad[i]) { free((void *)pa[i]); pa[i]=0; } // free good blocks
         else if (bad[i]) bad_nb++; // count bad blocks
  if (procid==0)
    printf("... good blocks freed (%d bad blocks left on proc0)\n", bad_nb);
  if (bad_nb && bad_nb!=1024) {
    for (;;) { char s1[8]/*,rc=0*/; /* wait for input within gnu-screen */
      if (!procid) system("date +%F\\ %T");
      printf("%d bad blocks detected at task%d=%s, sleep + fgets ...\n"
           " waiting for input (enter=sleep, char+ret=recheck_bad_blk):\n",
           bad_nb, procid, slurmd_nodename);
      fflush(stdout);
      #if 0  // do later if bad block found!? hpc18-problems?
        if ( mlockall(MCL_CURRENT) == -1 ) // 2020-07 lock bad block
          printf("ERROR: mlockall failed (non-root?)\n"); /* root only */
      #endif
      s1[0]='\0';if (!fgets(s1,8,stdin)) printf("ERROR fgets\n");
      printf("DBG: read char %04d %c\n", s1[0],
          ((s1[0]>='0' && s1[0]<='z')?s1[0]:'.'));
      if (s1[0]!='\0') { // simply just enter = "\n\0"
        for (bad_nb=i=0;i<nb;i++) if (pa[i]) bad[i]=0; /* reset counter */
        edac_blk_bitmap=0;
        for (ii=0; ii<n_loops;ii++) 
        /*rc=*/chkmem(pa,nb,num,ii,rnd,procid,bad,bad_nb);
        /* segfault? */
      } else sleep(10); // if(system("sleep 600"))break; // sleep forever
      // check every minute for read test of bad blk!? retrigger error
     // 2018-04-04 node168 2cpus 16*16GB-ECC-DIMMs ca. 30 ce-errors/13 loops
     //   block number is changing from run to run, constant within run
     // 2020-07 node016 bad block allocated + fget, further runs = ok
    }
  } // if bad_nb
  if (bad_nb==0 && edac_errors==0 && procid==0)
    printf("-- no error detected on procid=0 %s --\n", slurmd_nodename);
  if (bad_nb || edac_errors) // 2022-06 +bad_nb for srun -N 2-100
    // 11 edac errors needed to localize single bad block,
    //    unfortunately edac error comes delayed TODO blk-wise test?
    printf("ERR bad_nb= %3d ecc_errors=%d at node=%d %s (11 eccs needed)\n",
                bad_nb, edac_errors, procid, slurmd_nodename);
  if (!procid) system("date +%F\\ %T");
  // ToDo: how to integrate that into the jobsystem? extra-shared-queue?
  return 0;
}
